import pathlib
from typing import List

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import requests


def file2metadata(file: str) -> pd.DataFrame:
    """Read the metadata of a tortilla file given a URL. The
        server must support HTTP Range requests.

    Args:
        files (str): A URL pointing to the tortilla file.
    Returns:
        pd.DataFrame: The metadata of the tortilla file.
    """
    # Fetch the first 8 bytes of the file
    headers = {"Range": "bytes=0-50"}
    response: requests.Response = requests.get(file, headers=headers)
    static_bytes: bytes = response.content

    # SPLIT the static bytes
    MB: bytes = static_bytes[:2]
    FO: bytes = static_bytes[2:10]
    FL: bytes = static_bytes[10:18]
    DF: str = static_bytes[18:42].strip().decode()

    # Check if the file is a tortilla
    if MB != b"#y":
        raise ValueError("You are not a tortilla 🫓")

    # Interpret the bytes as a little-endian integer
    footer_offset: int = int.from_bytes(FO, "little")
    footer_length: int = int.from_bytes(FL, "little")

    # Fetch the footer
    headers = {"Range": f"bytes={footer_offset}-{footer_offset + footer_length}"}
    with requests.get(file, headers=headers) as response:

        # Interpret the response as a parquet table
        metadata = pq.read_table(pa.BufferReader(response.content)).to_pandas()

    # Add the file format and mode
    metadata["internal:file_format"] = DF
    metadata["internal:mode"] = "online"
    metadata["internal:subfile"] = metadata.apply(
        lambda row: f"/vsisubfile/{row['tortilla:offset']}_{row['tortilla:length']},/vsicurl/{file}",
        axis=1,
    )
    return metadata


def files2metadata(files: List[str]) -> pd.DataFrame:
    """Read the metadata of tortillas files given a set of URLs. The
        server must support HTTP Range requests.

    Args:
        files (List[str]): A list of URLs pointing to the
            tortilla files.

    Returns:
        pd.DataFrame: The metadata of the tortilla file.
    """

    container = []
    for file in files:

        # Fetch the first 8 bytes of the file
        headers = {"Range": "bytes=0-50"}
        response: requests.Response = requests.get(file, headers=headers)
        static_bytes: bytes = response.content

        # SPLIT the static bytes
        MB: bytes = static_bytes[:2]
        FO: bytes = static_bytes[2:10]
        FL: bytes = static_bytes[10:18]
        DF: str = static_bytes[18:42].strip().decode()

        # Check if the file is a tortilla
        if MB != b"#y":
            raise ValueError("You are not a tortilla 🫓")

        # Interpret the bytes as a little-endian integer
        footer_offset: int = int.from_bytes(FO, "little")
        footer_length: int = int.from_bytes(FL, "little")

        # Fetch the footer
        headers = {"Range": f"bytes={footer_offset}-{footer_offset + footer_length}"}
        with requests.get(file, headers=headers) as response:

            # Interpret the response as a parquet table
            metadata = pq.read_table(pa.BufferReader(response.content)).to_pandas()

        # Add the file format and mode
        metadata["internal:file_format"] = DF
        metadata["internal:mode"] = "online"
        metadata["internal:subfile"] = metadata.apply(
            lambda row: f"/vsisubfile/{row['tortilla:offset']}_{row['tortilla:length']},/vsicurl/{file}",
            axis=1,
        )
        container.append(metadata)

    return pd.concat(container, ignore_index=True)



import pathlib
from typing import List, Union

import pandas as pd
import geopandas as gpd
import shapely.wkt

from .load_local import file2metadata as local_file2metadata
from .load_local import files2metadata as local_files2metadata
from .load_remote import file2metadata as remote_file2metadata
from .load_remote import files2metadata as remote_files2metadata
from .utils import is_valid_url, snippet2files


def load(file: Union[str, pathlib.Path, List[pathlib.Path], List[str]]) -> pd.DataFrame:
    """Load the metadata of a tortilla file.

    Args:
        file (Union[str, pathlib.Path, List]): The path of
            the tortilla file. If the file is split into multiple
            parts, a list of paths is accepted. Also, multiple
            parts can be read by putting a asterisk (*) at the end
            of the file name. For example, "file*.tortilla". In this
            case, the function will create a list will all the partitions
            before the reading process.

    Returns:
        pd.DataFrame: The metadata of the tortilla file.
    """

    # Transform our snippet into a list of files
    # If it is not a snippet, it will return the same file
    file = snippet2files(file=file)

    if isinstance(file, list):
        if is_valid_url(file):
            metadata = remote_files2metadata(file)
        else:
            metadata = local_files2metadata(file)
    elif isinstance(file, (str, pathlib.Path)):
        if is_valid_url(file):
            metadata = remote_file2metadata(file)
        else:
            metadata = local_file2metadata(file)
    else:
        raise ValueError("Invalid file type. Must be a list, string or pathlib.Path.")

    # Convert the DataFrame to a GeoDataFrame
    if "stac:centroid" in metadata.columns:
        metadata = gpd.GeoDataFrame(
            data=metadata,
            geometry=metadata["stac:centroid"].apply(shapely.wkt.loads),
            crs="EPSG:4326"
        )

    # Sort the columns
    columns = metadata.columns
    prefixes = ["internal:", "tortilla:", "stac:", "rai:"]
    sorted_columns = [col for prefix in prefixes for col in columns if col.startswith(prefix)]
    rest = [col for col in columns if col not in sorted_columns and col != "geometry"]
    columns = sorted_columns + rest + (["geometry"] if "geometry" in columns else [])
    metadata = metadata[columns]

    return metadata



import pathlib
from typing import List, Union

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq


def file2metadata(file: Union[str, pathlib.Path]):
    """Read the metadata of tortilla file given a local path.

    Args:
        files (Union[str, pathlib.Path]): A local path pointing to the
            tortilla file.

    Returns:
        pd.DataFrame: The metadata of the tortilla file.
    """
    with open(file, "rb") as f:
        static_bytes = f.read(50)

        # SPLIT the static bytes
        MB: bytes = static_bytes[:2]
        FO: bytes = static_bytes[2:10]
        FL: bytes = static_bytes[10:18]
        DF: str = static_bytes[18:42].strip().decode()
        # DP: str = static_bytes[42:50]

        if MB != b"#y":
            raise ValueError("You are not a tortilla 🫓 or a TACO 🌮")

        # Read the NEXT 8 bytes of the file
        footer_offset: int = int.from_bytes(FO, "little")

        # Seek to the FOOTER offset
        f.seek(footer_offset)

        # Select the FOOTER length
        # Read the FOOTER
        footer_length: int = int.from_bytes(FL, "little")
        metadata = pq.read_table(pa.BufferReader(f.read(footer_length))).to_pandas()

        # Convert dataset to DataFrame
        metadata["internal:file_format"] = DF
        metadata["internal:mode"] = "local"
        metadata["internal:subfile"] = metadata.apply(
            lambda row: f"/vsisubfile/{row['tortilla:offset']}_{row['tortilla:length']},{file}",
            axis=1,
        )

    return metadata


def files2metadata(files: Union[List[str], List[pathlib.Path]]):
    """Read the metadata of tortilla files given local paths.

    Args:
        files (Union[List[str], List[pathlib.Path]]): A list of local
            paths pointing to the tortilla files.

    Returns:
        pd.DataFrame: The metadata of the tortilla file.
    """

    # Merge the metadata of the files
    container = []
    for file in files:
        with open(file, "rb") as f:
            static_bytes = f.read(50)

            # SPLIT the static bytes
            MB: bytes = static_bytes[:2]
            FO: bytes = static_bytes[2:10]
            FL: bytes = static_bytes[10:18]
            DF: str = static_bytes[18:42].strip().decode()
            # DP: str = static_bytes[42:50]

            if MB != b"#y":
                raise ValueError("You are not a tortilla 🫓 or a TACO 🌮")

            # Read the NEXT 8 bytes of the file
            footer_offset: int = int.from_bytes(FO, "little")

            # Seek to the FOOTER offset
            f.seek(footer_offset)

            # Select the FOOTER length
            # Read the FOOTER
            footer_length: int = int.from_bytes(FL, "little")
            metadata = pq.read_table(pa.BufferReader(f.read(footer_length))).to_pandas()

            # Convert dataset to DataFrame
            metadata["internal:file_format"] = DF
            metadata["internal:mode"] = "local"
            metadata["internal:subfile"] = metadata.apply(
                lambda row: f"/vsisubfile/{row['tortilla:offset']}_{row['tortilla:length']},{file}",
                axis=1,
            )
            container.append(metadata)

    return pd.concat(container, ignore_index=True)
